#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
import titus.P as P

# import special functions requred to compute distributions
from titus.lib.spec import logBetaFunction
from titus.lib.spec import incompleteBetaFunction
from titus.lib.spec import inverseIncompleteBetaFunction
from titus.lib.spec import regularizedGammaQ
from titus.lib.spec import regularizedGammaP
from titus.lib.spec import nChooseK

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "prob.dist."

class GaussianLL(LibFcn):
    name = prefix + "gaussianLL"
    sig = Sigs([Sig([{"x": P.Double()}, {"mu": P.Double()}, {"sigma": P.Double()}], P.Double()),
                Sig([{"x": P.Double()}, {"params": P.WildRecord("A", {"mean": P.Double(), "variance": P.Double()})}], P.Double())])
    errcodeBase = 13000
    def __call__(self, state, scope, pos, paramTypes, x, *others):
        if len(others) == 2:
            mu, sigma = others
        else:
            mu = others[0]["mean"]
            if math.isnan(others[0]["variance"]) or others[0]["variance"] < 0.0:
                raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
            else:
                sigma = math.sqrt(others[0]["variance"])
        if math.isinf(mu) or math.isnan(mu) or math.isinf(sigma) or math.isnan(sigma) or sigma < 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif sigma == 0.0:
            if x != mu:
                return float("-inf")
            else:
                return float("inf")
        else:
            return GaussianDistribution(mu, sigma, self.errcodeBase + 0, self.name, pos).LL(x)
provide(GaussianLL())

class GaussianCDF(LibFcn):
    name = prefix + "gaussianCDF"
    sig = Sigs([Sig([{"x": P.Double()}, {"mu": P.Double()}, {"sigma": P.Double()}], P.Double()),
                Sig([{"x": P.Double()}, {"params": P.WildRecord("A", {"mean": P.Double(), "variance": P.Double()})}], P.Double())])
    errcodeBase = 13010
    def __call__(self, state, scope, pos, paramTypes, x, *others):
        if len(others) == 2:
            mu, sigma = others
        else:
            mu = others[0]["mean"]
            if math.isnan(others[0]["variance"]) or others[0]["variance"] < 0.0:
                raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
            else:
                sigma = math.sqrt(others[0]["variance"])
        if math.isinf(mu) or math.isnan(mu) or math.isinf(sigma) or math.isnan(sigma) or sigma < 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif sigma == 0.0:
            if x < mu:
                return 0.0
            else:
                return 1.0
        else:
            return GaussianDistribution(mu, sigma, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(GaussianCDF())

# written using http://www.johndcook.com/normal_cdf_inverse.html
class GaussianQF(LibFcn):
    name = prefix + "gaussianQF"
    sig = Sigs([Sig([{"p": P.Double()}, {"mu": P.Double()}, {"sigma": P.Double()}], P.Double()),
                Sig([{"p": P.Double()}, {"params": P.WildRecord("A", {"mean": P.Double(), "variance": P.Double()})}], P.Double())])
    errcodeBase = 13020
    def __call__(self, state, scope, pos, paramTypes, p, *others):
        if len(others) == 2:
            mu, sigma = others
        else:
            mu = others[0]["mean"]
            if math.isnan(others[0]["variance"]) or others[0]["variance"] < 0.0:
                raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
            else:
                sigma = math.sqrt(others[0]["variance"])
        if math.isinf(mu) or math.isnan(mu) or math.isinf(sigma) or math.isnan(sigma) or sigma < 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 1.0:
            return float("inf")
        elif p == 0.0:
            return float("-inf")
        elif sigma == 0.0:
            return mu
        else:
            return GaussianDistribution(mu, sigma, self.errcodeBase + 0, self.name, pos).QF(p)
provide(GaussianQF())

################ Exponential
class ExponentialPDF(LibFcn):
    name = prefix + "exponentialPDF"
    sig = Sig([{"x": P.Double()}, {"lambda": P.Double()}], P.Double())
    errcodeBase = 13030
    def __call__(self, state, scope, pos, paramTypes, x, rate):
        if math.isinf(rate) or math.isnan(rate) or rate < 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif rate == 0.0:
            return 0.0
        elif x < 0.0:
            return 0.0
        elif x == 0.0:
            return rate
        else:
            return ExponentialDistribution(rate, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(ExponentialPDF())

class ExponentialCDF(LibFcn):
    name = prefix + "exponentialCDF"
    sig = Sig([{"x": P.Double()}, {"lambda": P.Double()}], P.Double())
    errcodeBase = 13040
    def __call__(self, state, scope, pos, paramTypes, x, rate):
        if math.isinf(rate) or math.isnan(rate) or rate < 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif rate == 0.0 or x == 0.0:
            return 0.0
        elif x <= 0.0:
            return 0.0
        else:
            return ExponentialDistribution(rate, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(ExponentialCDF())

class ExponentialQF(LibFcn):
    name = prefix + "exponentialQF"
    sig = Sig([{"p": P.Double()}, {"lambda": P.Double()}], P.Double())
    errcodeBase = 13050
    def __call__(self, state, scope, pos, paramTypes, p, rate):
        if math.isinf(rate) or math.isnan(rate) or rate < 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif rate == 0.0 and p == 0.0:
            return 0.0
        elif rate == 0.0 and p > 0:
            return float("inf")
        elif p == 1.0:
            return float("inf")
        elif p == 0.0:
            return 0.0
        else:
            return ExponentialDistribution(rate, self.errcodeBase + 0, self.name, pos).QF(p)
provide(ExponentialQF())

################ Chi2
class Chi2PDF(LibFcn):
    name = prefix + "chi2PDF"
    sig = Sig([{"x": P.Double()}, {"dof": P.Int()}], P.Double())
    errcodeBase = 13060
    def __call__(self, state, scope, pos, paramTypes, x, df):
        if df < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif df == 0:
            if x != 0:
                return 0.0
            else:
                return float("inf")
        elif x <= 0.0:
            return 0.0
        else:
            return Chi2Distribution(df, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(Chi2PDF())

class Chi2CDF(LibFcn):
    name = prefix + "chi2CDF"
    sig = Sig([{"x": P.Double()}, {"dof": P.Int()}], P.Double())
    errcodeBase = 13070
    def __call__(self, state, scope, pos, paramTypes, x, df):
        if df < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif df == 0:
            if x > 0:
                return 1.0
            else:
                return 0.0
        else:
            return Chi2Distribution(df, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(Chi2CDF())

class Chi2QF(LibFcn):
    name = prefix + "chi2QF"
    sig = Sig([{"p": P.Double()}, {"dof": P.Int()}], P.Double())
    errcodeBase = 13080
    def __call__(self, state, scope, pos, paramTypes, p, df):
        if df < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 1.0:
            return float("inf")
        elif df == 0:
            return 0.0
        elif p == 0.0:
            return 0.0
        else:
            return Chi2Distribution(df, self.errcodeBase + 0, self.name, pos).QF(p)
provide(Chi2QF())

################ Poisson #######################################
class PoissonPDF(LibFcn):
    name = prefix + "poissonPDF"
    sig = Sig([{"x": P.Int()}, {"lambda": P.Double()}], P.Double())
    errcodeBase = 13090
    def __call__(self, state, scope, pos, paramTypes, x, lamda):
        if math.isinf(lamda) or math.isnan(lamda) or lamda < 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif lamda == 0:
            if x != 0:
                return 0.0
            else:
                return 1.0
        elif x < 0:
            return 0.0
        else:
            return PoissonDistribution(lamda, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(PoissonPDF())

class PoissonCDF(LibFcn):
    name = prefix + "poissonCDF"
    sig = Sig([{"x": P.Int()}, {"lambda": P.Double()}], P.Double())
    errcodeBase = 13100
    def __call__(self, state, scope, pos, paramTypes, x, lamda):
        if math.isinf(lamda) or math.isnan(lamda) or lamda < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif lamda == 0:
            if x >= 0:
                return 1.0
            else:
                return 0.0
        else:
            return PoissonDistribution(lamda, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(PoissonCDF())

class PoissonQF(LibFcn):
    name = prefix + "poissonQF"
    sig = Sig([{"p": P.Double()}, {"lambda": P.Double()}], P.Double())
    errcodeBase = 13110
    def __call__(self, state, scope, pos, paramTypes, p, lamda):
        if math.isinf(lamda) or math.isnan(lamda) or lamda < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif lamda == 0:
            return 0.0
        elif p == 1:
            return float("inf")
        elif p == 0:
            return 0.0
        else:
            return PoissonDistribution(lamda, self.errcodeBase + 0, self.name, pos).QF(p)
provide(PoissonQF())

################ Gamma
class GammaPDF(LibFcn):
    name = prefix + "gammaPDF"
    sig = Sig([{"x": P.Double()}, {"shape": P.Double()}, {"scale": P.Double()}], P.Double())
    errcodeBase = 13120
    def __call__(self, state, scope, pos, paramTypes, x, shape, scale):
        if math.isinf(shape) or math.isnan(shape) or math.isinf(scale) or math.isnan(scale) or shape < 0 or scale < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif shape == 0 or scale == 0:
            if x != 0:
                return 0.0
            else:
                return float("inf")
        elif x < 0:
            return 0.0
        elif x == 0:
            if shape < 1:
                return float("inf")
            elif shape == 1:
                return 1.0/scale
            else:
                return 0.0
        else:
            return GammaDistribution(shape, scale, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(GammaPDF())

class GammaCDF(LibFcn):
    name = prefix + "gammaCDF"
    sig = Sig([{"x": P.Double()}, {"shape": P.Double()}, {"scale": P.Double()}], P.Double())
    errcodeBase = 13130
    def __call__(self, state, scope, pos, paramTypes, x, shape, scale):
        if math.isinf(shape) or math.isnan(shape) or math.isinf(scale) or math.isnan(scale) or shape < 0 or scale < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif shape == 0 or scale == 0:
            if x != 0:
                return 1.0
            else:
                return 0.0
        elif x < 0:
            return 0.0
        else:
            return GammaDistribution(shape, scale, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(GammaCDF())

class GammaQF(LibFcn):
    name = prefix + "gammaQF"
    sig = Sig([{"p": P.Double()}, {"shape": P.Double()}, {"scale": P.Double()}], P.Double())
    errcodeBase = 13140
    def __call__(self, state, scope, pos, paramTypes, p, shape, scale):
        if math.isinf(shape) or math.isnan(shape) or math.isinf(scale) or math.isnan(scale) or shape <= 0 or scale <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 1.0:
            return float("inf")
        elif p == 0.0:
            return 0.0
        else:
            return GammaDistribution(shape, scale, self.errcodeBase + 0, self.name, pos).QF(p)
provide(GammaQF())

################ Beta
class BetaPDF(LibFcn):
    name = prefix + "betaPDF"
    sig = Sig([{"x": P.Double()}, {"a": P.Double()}, {"b": P.Double()}], P.Double())
    errcodeBase = 13150
    def __call__(self, state, scope, pos, paramTypes, x, shape1, shape2):
        if math.isinf(shape1) or math.isnan(shape1) or math.isinf(shape2) or math.isnan(shape2) or shape1 <= 0 or shape2 <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x <= 0 or x >= 1:
            return 0.0
        else:
            return BetaDistribution(shape1, shape2, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(BetaPDF())

class BetaCDF(LibFcn):
    name = prefix + "betaCDF"
    sig = Sig([{"x": P.Double()}, {"a": P.Double()}, {"b": P.Double()}], P.Double())
    errcodeBase = 13160
    def __call__(self, state, scope, pos, paramTypes, x, shape1, shape2):
        if math.isinf(shape1) or math.isnan(shape1) or math.isinf(shape2) or math.isnan(shape2) or shape1 <= 0 or shape2 <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x <= 0:
            return 0.0
        elif x >= 1:
            return 1.0
        else:
            return BetaDistribution(shape1, shape2, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(BetaCDF())

class BetaQF(LibFcn):
    name = prefix + "betaQF"
    sig = Sig([{"p": P.Double()}, {"a": P.Double()}, {"b": P.Double()}], P.Double())
    errcodeBase = 13170
    def __call__(self, state, scope, pos, paramTypes, p, shape1, shape2):
        if math.isinf(shape1) or math.isnan(shape1) or math.isinf(shape2) or math.isnan(shape2) or shape1 <= 0 or shape2 <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 1:
            return 1.0
        elif p == 0:
            return 0.0
        else:
            return BetaDistribution(shape1, shape2, self.errcodeBase + 0, self.name, pos).QF(p)
provide(BetaQF())

################ Cauchy
class CauchyPDF(LibFcn):
    name = prefix + "cauchyPDF"
    sig = Sig([{"x": P.Double()}, {"location": P.Double()}, {"scale": P.Double()}], P.Double())
    errcodeBase = 13180
    def __call__(self, state, scope, pos, paramTypes, x, location, scale):
        if math.isinf(location) or math.isnan(location) or math.isinf(scale) or math.isnan(scale) or scale <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return CauchyDistribution(location, scale, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(CauchyPDF())

class CauchyCDF(LibFcn):
    name = prefix + "cauchyCDF"
    sig = Sig([{"x": P.Double()}, {"location": P.Double()}, {"scale": P.Double()}], P.Double())
    errcodeBase = 13190
    def __call__(self, state, scope, pos, paramTypes, x, location, scale):
        if math.isinf(location) or math.isnan(location) or math.isinf(scale) or math.isnan(scale) or scale <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return CauchyDistribution(location, scale, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(CauchyCDF())

class CauchyQF(LibFcn):
    name = prefix + "cauchyQF"
    sig = Sig([{"p": P.Double()}, {"location": P.Double()}, {"scale": P.Double()}], P.Double())
    errcodeBase = 13200
    def __call__(self, state, scope, pos, paramTypes, p, location, scale):
        if math.isinf(location) or math.isnan(location) or math.isinf(scale) or math.isnan(scale) or scale <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 1:
            return float("inf")
        elif p == 0:
            return float("-inf")
        else:
            return CauchyDistribution(location, scale, self.errcodeBase + 0, self.name, pos).QF(p)
provide(CauchyQF())

################ F
class FPDF(LibFcn):
    name = prefix + "fPDF"
    sig = Sig([{"x": P.Double()}, {"d1": P.Int()}, {"d2": P.Int()}], P.Double())
    errcodeBase = 13210
    def __call__(self, state, scope, pos, paramTypes, x, d1, d2):
        if d2 <= 0 or d1 <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x <= 0:
            return 0.0
        else:
            return FDistribution(d1, d2, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(FPDF())

class FCDF(LibFcn):
    name = prefix + "fCDF"
    sig = Sig([{"x": P.Double()}, {"d1": P.Int()}, {"d2": P.Int()}], P.Double())
    errcodeBase = 13220
    def __call__(self, state, scope, pos, paramTypes, x, d1, d2):
        if d2 <= 0 or d1 <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return FDistribution(d1, d2, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(FCDF())

class FQF(LibFcn):
    name = prefix + "fQF"
    sig = Sig([{"p": P.Double()}, {"d1": P.Int()}, {"d2": P.Int()}], P.Double())
    errcodeBase = 13230
    def __call__(self, state, scope, pos, paramTypes, p, d1, d2):
        if d1 <= 0 or d2 <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 1:
            return float("inf")
        else:
            return FDistribution(d1, d2, self.errcodeBase + 0, self.name, pos).QF(p)
provide(FQF())

################ Lognormal
class LognormalPDF(LibFcn):
    name = prefix + "lognormalPDF"
    sig = Sig([{"x": P.Double()}, {"meanlog": P.Double()}, {"sdlog": P.Double()}], P.Double())
    errcodeBase = 13240
    def __call__(self, state, scope, pos, paramTypes, x, meanlog, sdlog):
        if math.isinf(meanlog) or math.isnan(meanlog) or math.isinf(sdlog) or math.isnan(sdlog) or sdlog <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return LognormalDistribution(meanlog, sdlog, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(LognormalPDF())

class LognormalCDF(LibFcn):
    name = prefix + "lognormalCDF"
    sig = Sig([{"x": P.Double()}, {"meanlog": P.Double()}, {"sdlog": P.Double()}], P.Double())
    errcodeBase = 13250
    def __call__(self, state, scope, pos, paramTypes, x, meanlog, sdlog):
        if math.isinf(meanlog) or math.isnan(meanlog) or math.isinf(sdlog) or math.isnan(sdlog) or sdlog <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return LognormalDistribution(meanlog, sdlog, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(LognormalCDF())

class LognormalQF(LibFcn):
    name = prefix + "lognormalQF"
    sig = Sig([{"p": P.Double()}, {"meanlog": P.Double()}, {"sdlog": P.Double()}], P.Double())
    errcodeBase = 13260
    def __call__(self, state, scope, pos, paramTypes, p, meanlog, sdlog):
        if math.isinf(meanlog) or math.isnan(meanlog) or math.isinf(sdlog) or math.isnan(sdlog) or sdlog <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 1:
            return float("inf")
        else:
            return LognormalDistribution(meanlog, sdlog, self.errcodeBase + 0, self.name, pos).QF(p)
provide(LognormalQF())

################ T
class TPDF(LibFcn):
    name = prefix + "tPDF"
    sig = Sig([{"x": P.Double()}, {"dof": P.Int()}], P.Double())
    errcodeBase = 13270
    def __call__(self, state, scope, pos, paramTypes, x, df):
        if df <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return TDistribution(df, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(TPDF())

class TCDF(LibFcn):
    name = prefix + "tCDF"
    sig = Sig([{"x": P.Double()}, {"dof": P.Int()}], P.Double())
    errcodeBase = 13280
    def __call__(self, state, scope, pos, paramTypes, x, df):
        if df <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return TDistribution(df, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(TCDF())

class TQF(LibFcn):
    name = prefix + "tQF"
    sig = Sig([{"p": P.Double()}, {"dof": P.Int()}], P.Double())
    errcodeBase = 13290
    def __call__(self, state, scope, pos, paramTypes, p, df):
        if df <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 1:
            return float("inf")
        elif p == 0:
            return float("-inf")
        else:
            return TDistribution(df, self.errcodeBase + 0, self.name, pos).QF(p)
provide(TQF())

################ Binomial
class BinomialPDF(LibFcn):
    name = prefix + "binomialPDF"
    sig = Sig([{"x": P.Int()}, {"size": P.Int()}, {"prob": P.Double()}], P.Double())
    errcodeBase = 13300
    def __call__(self, state, scope, pos, paramTypes, x, size, prob):
        if math.isinf(prob) or math.isnan(prob) or size <= 0 or prob < 0 or prob > 1:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x < 0:
            return 0.0
        elif x >= size:
            return 0.0
        else:
            return BinomialDistribution(size, prob, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(BinomialPDF())

class BinomialCDF(LibFcn):
    name = prefix + "binomialCDF"
    sig = Sig([{"x": P.Double()}, {"size": P.Int()}, {"prob": P.Double()}], P.Double())
    errcodeBase = 13310
    def __call__(self, state, scope, pos, paramTypes, x, size, prob):
        if math.isinf(prob) or math.isnan(prob) or size <= 0 or prob < 0 or prob > 1:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x < 0:
            return 0.0
        elif x >= size:
            return 1.0
        elif prob == 1:
            if x < size:
                return 0.0
            else:
                return 1.0
        elif prob == 0:
            return 1.0
        else:
            return BinomialDistribution(size, prob, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(BinomialCDF())

class BinomialQF(LibFcn):
    name = prefix + "binomialQF"
    sig = Sig([{"p": P.Double()}, {"size": P.Int()}, {"prob": P.Double()}], P.Double())
    errcodeBase = 13320
    def __call__(self, state, scope, pos, paramTypes, p, size, prob):
        if math.isinf(prob) or math.isnan(prob) or size <= 0 or prob < 0 or prob > 1:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 1:
            return float(size)
        elif p == 0:
            return 0.0
        else:
            return BinomialDistribution(size, prob, self.errcodeBase + 0, self.name, pos).QF(p)
provide(BinomialQF())

################ Uniform
class UniformPDF(LibFcn):
    name = prefix + "uniformPDF"
    sig = Sig([{"x": P.Double()}, {"min": P.Double()}, {"max": P.Double()}], P.Double())
    errcodeBase = 13330
    def __call__(self, state, scope, pos, paramTypes, x, min, max):
        if math.isinf(min) or math.isnan(min) or math.isinf(max) or math.isnan(max) or min >= max:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        return UniformDistribution(min, max, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(UniformPDF())

class UniformCDF(LibFcn):
    name = prefix + "uniformCDF"
    sig = Sig([{"x": P.Double()}, {"min": P.Double()}, {"max": P.Double()}], P.Double())
    errcodeBase = 13340
    def __call__(self, state, scope, pos, paramTypes, x, min, max):
        if math.isinf(min) or math.isnan(min) or math.isinf(max) or math.isnan(max) or min >= max:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        return UniformDistribution(min, max, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(UniformCDF())

class UniformQF(LibFcn):
    name = prefix + "uniformQF"
    sig = Sig([{"p": P.Double()}, {"min": P.Double()}, {"max": P.Double()}], P.Double())
    errcodeBase = 13350
    def __call__(self, state, scope, pos, paramTypes, p, min, max):
        if math.isinf(min) or math.isnan(min) or math.isinf(max) or math.isnan(max) or min >= max:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return UniformDistribution(min, max, self.errcodeBase + 0, self.name, pos).QF(p)
provide(UniformQF())

################ Geometric
class GeometricPDF(LibFcn):
    name = prefix + "geometricPDF"
    sig = Sig([{"x": P.Int()}, {"prob": P.Double()}], P.Double())
    errcodeBase = 13360
    def __call__(self, state, scope, pos, paramTypes, x, prob):
        if math.isinf(prob) or math.isnan(prob) or prob <= 0 or prob > 1:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return GeometricDistribution(prob, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(GeometricPDF())

class GeometricCDF(LibFcn):
    name = prefix + "geometricCDF"
    sig = Sig([{"x": P.Double()}, {"prob": P.Double()}], P.Double())
    errcodeBase = 13370
    def __call__(self, state, scope, pos, paramTypes, x, prob):
        if math.isinf(prob) or math.isnan(prob) or prob <= 0 or prob > 1:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return GeometricDistribution(prob, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(GeometricCDF())

class GeometricQF(LibFcn):
    name = prefix + "geometricQF"
    sig = Sig([{"p": P.Double()}, {"prob": P.Double()}], P.Double())
    errcodeBase = 13380
    def __call__(self, state, scope, pos, paramTypes, p, prob):
        if math.isinf(prob) or math.isnan(prob) or prob <= 0 or prob > 1:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 1:
            return float("inf")
        else:
            return GeometricDistribution(prob, self.errcodeBase + 0, self.name, pos).QF(p)
provide(GeometricQF())

################ Hypergeometric
class HypergeometricPDF(LibFcn):
    name = prefix + "hypergeometricPDF"
    sig = Sig([{"x": P.Int()}, {"m": P.Int()}, {"n": P.Int()}, {"k": P.Int()}], P.Double())
    errcodeBase = 13390
    def __call__(self, state, scope, pos, paramTypes, x, m, n, k):
        if m + n < k or m < 0 or n <= 0 or m + n == 0 or k < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x > m:
            return 0.0
        else:
            return HypergeometricDistribution(m, n, k, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(HypergeometricPDF())

class HypergeometricCDF(LibFcn):
    name = prefix + "hypergeometricCDF"
    sig = Sig([{"x": P.Int()}, {"m": P.Int()}, {"n": P.Int()}, {"k": P.Int()}], P.Double())
    errcodeBase = 13400
    def __call__(self, state, scope, pos, paramTypes, x, m, n, k):
        if m + n < k or m < 0 or n <= 0 or m + n == 0 or k < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x > m:
            return 0.0
        else:
            return HypergeometricDistribution(m, n, k, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(HypergeometricCDF())

class HypergeometricQF(LibFcn):
    name = prefix + "hypergeometricQF"
    sig = Sig([{"p": P.Double()}, {"m": P.Int()}, {"n": P.Int()}, {"k": P.Int()}], P.Double())
    errcodeBase = 13410
    def __call__(self, state, scope, pos, paramTypes, p, m, n, k):
        if m + n < k or m < 0 or n <= 0 or m + n == 0 or k < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return HypergeometricDistribution(m, n, k, self.errcodeBase + 0, self.name, pos).QF(p)
provide(HypergeometricQF())

################ Weibull
class WeibullPDF(LibFcn):
    name = prefix + "weibullPDF"
    sig = Sig([{"x": P.Double()}, {"shape": P.Double()}, {"scale": P.Double()}], P.Double())
    errcodeBase = 13420
    def __call__(self, state, scope, pos, paramTypes, x, shape, scale):
        if math.isinf(shape) or math.isnan(shape) or math.isinf(scale) or math.isnan(scale) or shape <= 0 or scale <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x >= 0:
            return WeibullDistribution(shape, scale, self.errcodeBase + 0, self.name, pos).PDF(x)
        else:
            return 0.0
provide(WeibullPDF())

class WeibullCDF(LibFcn):
    name = prefix + "weibullCDF"
    sig = Sig([{"x": P.Double()}, {"shape": P.Double()}, {"scale": P.Double()}], P.Double())
    errcodeBase = 13430
    def __call__(self, state, scope, pos, paramTypes, x, shape, scale):
        if math.isinf(shape) or math.isnan(shape) or math.isinf(scale) or math.isnan(scale) or shape <= 0 or scale <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x < 0:
            return 0.0
        else:
            return WeibullDistribution(shape, scale, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(WeibullCDF())

class WeibullQF(LibFcn):
    name = prefix + "weibullQF"
    sig = Sig([{"p": P.Double()}, {"shape": P.Double()}, {"scale": P.Double()}], P.Double())
    errcodeBase = 13440
    def __call__(self, state, scope, pos, paramTypes, p, shape, scale):
        if math.isinf(shape) or math.isnan(shape) or math.isinf(scale) or math.isnan(scale) or shape <= 0 or scale <= 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        else:
            return WeibullDistribution(shape, scale, self.errcodeBase + 0, self.name, pos).QF(p)
provide(WeibullQF())

################ NegativeBinomial
class NegativeBinomialPDF(LibFcn):
    name = prefix + "negativeBinomialPDF"
    sig = Sig([{"x": P.Int()}, {"size": P.Int()}, {"prob": P.Double()}], P.Double())
    errcodeBase = 13450
    def __call__(self, state, scope, pos, paramTypes, x, size, prob):
        if math.isinf(prob) or math.isnan(prob) or prob > 1 or prob <= 0 or size < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x == 0 and size == 0:
            return 1.0
        elif size == 0:
            return 0.0
        else:
            return NegativeBinomialDistribution(size, prob, self.errcodeBase + 0, self.name, pos).PDF(x)
provide(NegativeBinomialPDF())

class NegativeBinomialCDF(LibFcn):
    name = prefix + "negativeBinomialCDF"
    sig = Sig([{"x": P.Double()}, {"size": P.Int()}, {"prob": P.Double()}], P.Double())
    errcodeBase = 13460
    def __call__(self, state, scope, pos, paramTypes, x, size, prob):
        if math.isinf(prob) or math.isnan(prob) or prob > 1 or prob <= 0 or size < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif math.isinf(x) or math.isnan(x):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif x == 0 and size == 0:
            return 1.0
        elif size == 0:
            return 0.0
        elif x < 0:
            return 0.0
        else:
            return NegativeBinomialDistribution(size, prob, self.errcodeBase + 0, self.name, pos).CDF(x)
provide(NegativeBinomialCDF())

class NegativeBinomialQF(LibFcn):
    name = prefix + "negativeBinomialQF"
    sig = Sig([{"p": P.Double()}, {"size": P.Int()}, {"prob": P.Double()}], P.Double())
    errcodeBase = 13470
    def __call__(self, state, scope, pos, paramTypes, p, size, prob):
        if math.isinf(prob) or math.isnan(prob) or prob <= 0 or prob > 1 or size < 0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, pos)
        elif not (0.0 <= p <= 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, pos)
        elif p == 0 and size == 0:
            return 0.0
        elif size == 0:
            return 0.0
        elif p == 1:
            return float("inf")
        else:
            return NegativeBinomialDistribution(size, prob, self.errcodeBase + 0, self.name, pos).QF(p)
provide(NegativeBinomialQF())

#########################################################################################
##### The actual distribution functions #################################################
#########################################################################################

################### Gaussian
class GaussianDistribution(object):
    def __init__(self, mu, sigma, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        if sigma < 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)
        self.mu = mu
        self.sigma = sigma

    def LL(self, x):
        if (self.sigma == 0.0) and (x == self.mu):
            return float("inf")
        elif (self.sigma == 0.0) and (x != self.mu):
            return float("-inf")
        else:
            term1 = -(x - self.mu)**2/(2.0 * self.sigma**2)
            term2 = math.log(self.sigma * math.sqrt(2.0 * math.pi))
            return term1 - term2

    def CDF(self, x):
        if (self.sigma == 0.0) and (x < self.mu):
            return 0.0
        elif (self.sigma == 0.0) and (x >= self.mu):
            return 1.0
        else:
            return 0.5 * (1.0 + math.erf((x - self.mu)/(self.sigma * math.sqrt(2.0))))

    def QF(self, p):
        if (p > 1.0) or (p < 0.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p == 1.0):
            return float("inf")
        elif (p == 0.0):
            return float("-inf")
        elif (self.sigma == 0.0):
            return self.mu
        else:
            # http://www.johnkerl.org/python/sp_funcs_m.py.txt
            c0 = 2.515517
            c1 = 0.802853
            c2 = 0.010328
            d1 = 1.432788
            d2 = 0.189269
            d3 = 0.001308

            sign = -1.0
            if (p > 0.5):
                sign = 1.0
                p = 1.0 - p

            arg = -2.0*math.log(p)
            t = math.sqrt(arg)
            g = t - (c0 + t*(c1 + t*c2)) / (1.0 + t*(d1 + t*(d2 + t*d3)))
            standard_normal_qf = sign*g
            return self.mu + self.sigma*standard_normal_qf

################### Exponential
class ExponentialDistribution(object):
    def __init__(self, rate, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.rate = rate
        if (self.rate < 0.0):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        if (self.rate == 0.0):
            return 0.0
        elif (x < 0.0):
            return 0.0
        elif (x == 0.0):
            return self.rate
        else:
            return self.rate*math.exp(-self.rate*x)

    def CDF(self, x):
        if (self.rate == 0.0) or (x == 0.0):
            return 0.0
        elif (x <= 0.0):
            return 0.0
        else:
            return 1.0 - math.exp(-self.rate*x)

    def QF(self, p):
        if (p > 1.0) or (p < 0.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (self.rate == 0.0) and (p == 0.0):
            return 0.0
        elif (self.rate == 0.0) and (p > 0):
            return float("inf")
        elif (p == 1.0):
            return float("inf")
        elif (p == 0.0):
            return 0.0
        else:
            return -math.log(1.0 - p)/self.rate

################### Chi2
# from: http://www.stat.tamu.edu/~jnewton/604/chap3.pdf
class Chi2Distribution(object):
    def __init__(self, DOF, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.maxIter = 200
        self.epsilon = 1e-15
        self.DOF     = DOF
        if (self.DOF < 0):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self,x):
        if (self.DOF == 0) and (x != 0.0):
            return 0.0
        elif (self.DOF == 0) and (x == 0.0):
            return float("inf")
        elif (x == 0.0) and (self.DOF < 2):
            return float("inf")
        elif (x < 0.0):
            return 0.0
        else:
            return GammaDistribution(self.DOF/2.0, 2.0, self.name, self.errcodeBase, self.pos).PDF(x)

    def CDF(self,x):
        if math.isnan(x):
            return float("nan")
        elif (self.DOF == 0) and (x > 0.0):
            return 1.0
        elif (self.DOF == 0) and (x <= 0.0):
            return 0.0
        elif (x <= 0.0):
            return 0.0
        else:
            return GammaDistribution(self.DOF/2.0, 2.0, self.name, self.errcodeBase, self.pos).CDF(x)

    def QF(self,p):
        if (p > 1.0) or (p < 0.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p == 1.0):
            return float("inf")
        elif (self.DOF == 0):
            return 0.0
        elif (p == 0.0):
            return 0.0
        else:
            return GammaDistribution(self.DOF/2.0, 2.0, self.name, self.errcodeBase, self.pos).QF(p)

################### Poisson
class PoissonDistribution(object):
    def __init__(self, lamda, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.lamda = float(lamda)
        if (self.lamda < 0.0):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        if (self.lamda == 0.0):
            if (x != 0.0):
                return 0.0
            else:
                return 1.0
        elif (x < 0.0):
            return 0.0
        else:
            return math.exp(x * math.log(self.lamda) - self.lamda - math.lgamma(x + 1.0))

    def CDF(self, x):
        if math.isnan(x):
            return float("nan")
        elif (self.lamda == 0.0):
            if (x >= 0.0):
                return 1.0
            else:
                return 0.0
        elif (x >= 0.0):
            return regularizedGammaQ(math.floor(x + 1.0), self.lamda)
        else:
            return 0.0

    def QF(self, p):
        if math.isnan(p):
            return float("nan")
        elif (self.lamda == 0.0):
            return 0.0
        elif (p > 1.0) or (p < 0.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p == 1.0):
            return float("inf")
        elif (p == 0.0):
            return 0.0
        else:
            # step through CDFs until we find the right one
            x = 0
            p0 = 0.0
            while (p0 <= p):
                p0 = self.CDF(x)
                x += 1
            return x - 1

################### Gamma
class GammaDistribution(object):
    def __init__(self, shape, scale, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        # alpha: shape parameter
        self.alpha   = shape
        # beta:  scale parameter
        self.beta    = scale
        self.epsilon = 1e-15
        self.maxIter = 800

        if (self.alpha < 0.0) or (self.beta < 0.0):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        if (self.alpha == 0.0) or (self.beta == 0.0):
            if (x != 0.0):
                return 0.0
            else:
                return float("inf")
        elif (x < 0.0):
            return 0.0
        elif (self.alpha == 1.0) and (x == 0.0):
            return self.beta
        else:
            try:
                term1a = math.log(x/self.beta) * (self.alpha - 1.0)
            except ValueError:
                term1a = float("-inf") * (self.alpha - 1.0)
            term1 = term1a - math.log(self.beta)
            term2 = -x/self.beta
            term3 = math.lgamma(self.alpha)
            return math.exp(term1 + (term2 - term3))

    def CDF(self, x):
        if (self.alpha == 0.0) or (self.beta == 0.0):
            if (x != 0.0):
                return 1.0
            else:
                return 0.0
        elif (x <= 0.0):
            return 0.0
        else:
            return regularizedGammaP(self.alpha, x/self.beta)

    def QF(self,p):
        if (p > 1.0) or (p < 0.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p == 1.0):
            return float("inf")
        elif (p == 0.0):
            return 0.0
        else:
            y = self.alpha*self.beta
            y_old = y
            for i in range(0,100):
                h = (self.CDF(y_old) - p)/self.PDF(y_old)
                if y_old - h <= 0.0:
                    y_new = y_old / 2.0
                else:
                    y_new = y_old - h
                if abs(y_new) <= self.epsilon:
                    y_new = y_old/10.0
                    h = y_old - y_new
                if abs(h) < math.sqrt(self.epsilon):
                    break
                y_old = y_new
            return y_new

################### Beta
class BetaDistribution(object):
    def __init__(self, alpha, beta, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        # first shape parameter
        self.alpha = alpha
        # second shape parameter
        self.beta  = beta
        if (self.alpha <= 0.0) or (self.beta <= 0.0):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)
        # normalization factor
        self.Z = math.lgamma(self.alpha) + math.lgamma(self.beta) \
                 - math.lgamma(self.alpha + self.beta)
        # tolerance
        self.epsilon = 1e-15
        # max Iterations
        self.maxIter = 1000

    def PDF(self,x):
        if (x <= 0.0) or (x >= 1.0):
            return 0.0
        else:
            logX = math.log(x)
            if (x < 0.0) and (x > 0.0):
                return 0.0
            log1mX = math.log1p(-x)
            ret = math.exp((self.alpha - 1.0) * logX + (self.beta - 1.0) \
                  * log1mX - self.Z)
            return ret

    def CDF(self,x):
        if math.isnan(x):
            return float("nan")
        elif (x <= 0.0):
            return 0.0
        elif (x >= 1.0):
            return 1.0
        else:
            return incompleteBetaFunction(x,self.alpha,self.beta)

    def QF(self,p):
        if math.isnan(p):
            return float("nan")
        elif (p > 1.0) or (p < 0.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p == 1.0):
            return 1.0
        elif (p == 0.0):
            return 0.0
        else:
            return inverseIncompleteBetaFunction(p,self.alpha,self.beta)

################### Cauchy
class CauchyDistribution(object):
    def __init__(self, location, scale, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.loc = location
        self.s = scale
        if self.s <= 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        term1 = 1.0/(math.pi*self.s)
        term2 = 1.0/(1 + pow((x - self.loc)/self.s,2))
        return term1 * term2

    def CDF(self, x):
        return 0.5 + math.atan2(x-self.loc, self.s)*(1.0/math.pi)

    def QF(self, p):
        if (p < 0.0) or (p > 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p == 1.0):
            return float("inf")
        elif (p == 0.0):
            return float("-inf")
        else:
            return self.loc + self.s*math.tan(math.pi*(p - 0.5))

################### F
# from: http://www.stat.tamu.edu/~jnewton/604/chap3.pdf
class FDistribution(object):
    def __init__(self, upperDOF, lowerDOF, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.maxIter = 1000
        self.epsilon = 1e-8
        self.d1 = float(upperDOF)
        self.d2 = float(lowerDOF)
        if (self.d1 <= 0.0) or (self.d2 <= 0.0):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self,x):
        if (x <= 0.0):
            return 0.0
        elif (x == 0) and (self.d1 < 2.0):
            return float("inf")
        elif (x == 0) and (self.d1 == 2.0):
            return 1.0
        else:
            num_arg1 = pow(self.d1/self.d2, self.d1/2.0)
            num_arg2 = pow(x, (self.d1/2.0)-1.0)
            den_arg1 = math.exp(logBetaFunction(self.d1/2.0, self.d2/2.0))
            den_arg2 = pow((1.0 + (self.d1*x)/self.d2), (self.d1 + self.d2)/2.0)
            return (num_arg1*num_arg2)/(den_arg1*den_arg2)

    def CDF(self,x):
        if math.isnan(x):
            return float("nan")
        elif x <= 0.0:
            return 0.0
        else:
            arg1 = (self.d1*x)/(self.d1*x + self.d2)
            arg2 = self.d1/2.0
            arg3 = self.d2/2.0
            return incompleteBetaFunction(arg1, arg2, arg3)

    def QF(self, p):
        if (p < 0.0) or (p > 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p == 1.0):
            return float("inf")
        elif (p == 0.0):
            return 0.0
        else:
            low = 0.0
            high = 1.0
            while self.CDF(high) < p:
                high *= 2.0
            diff = None
            while diff is None or abs(diff) > self.epsilon:
                mid = (low + high) / 2.0
                diff = self.CDF(mid) - p
                if diff > 0:
                    high = mid
                else:
                    low = mid
            return mid

################### Lognormal
class LognormalDistribution(object):
    def __init__(self, meanlog, sdlog, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.mu = meanlog
        self.sigma = sdlog
        self.epsilon = 1e-8
        self.maxIter = 100
        if self.sigma <= 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        if x <= 0.0:
            return 0.0
        else:
            term1 = 1.0/(x*self.sigma*math.sqrt(2.0*math.pi))
            term2 = pow(math.log(x) - self.mu, 2.0)/(2.0*pow(self.sigma, 2.0))
            return term1 * math.exp(-term2)

    def CDF(self, x):
        if x <= 0.0:
            return 0.0
        else:
            return GaussianDistribution(0.0, 1.0, self.name, self.errcodeBase, self.pos).CDF((math.log(x) - self.mu)/self.sigma)

    def QF(self, p):
        if math.isnan(p):
            return float("nan")
        if (p < 0.0) or (p > 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p == 0.0):
            return 0.0
        elif (p == 1.0):
            return float("inf")
        else:
            low = 0.0
            high = 1.0
            while self.CDF(high) < p:
                high *= 2.0
            diff = None
            while diff is None or abs(diff) > self.epsilon:
                mid = (low + high) / 2.0
                diff = self.CDF(mid) - p
                if diff > 0:
                    high = mid
                else:
                    low = mid
            return mid

            # # Using Newton-Raphson algorithm
            # if p <= .001:
            #     self.epsilon = 1e-5
            # p1 = p
            # if (p1 > 0.8) and (p1 < 0.9):
            #     p2 = .5
            # else:
            #     p2 = 0.85
            # counter = 0
            # while (abs(p1 - p2) > self.epsilon) and (counter < self.maxIter):
            #     q2 = (self.CDF(p2) - p)
            #     p1 = p2
            #     p2 = p1 - (q2/self.PDF(p1))
            #     counter += 1
            # return p2

################### Student's T
class TDistribution(object):
    def __init__(self, DOF, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.df = float(DOF)
        self.epsilon = 1e-8
        self.maxIter = 800
        if self.df <= 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        term1 = 1.0/(math.sqrt(self.df) * math.exp(logBetaFunction(0.5, self.df/2.0)))
        term2 = pow(1.0 + (x*x/self.df), -(self.df + 1.0)/2.0)
        return term1 * term2

    def CDF(self, x):
        if math.isnan(x):
            return float("nan")
        arg1 = self.df/(self.df + x*x)
        arg2 = self.df/2.0
        arg3 = 0.5
        if (x > 0):
            return 1.0 - 0.5*incompleteBetaFunction(arg1, arg2, arg3)
        elif (x == 0.0):
            return 0.5
        else:
            return 0.5*incompleteBetaFunction(arg1, arg2, arg3)

    def QF(self, p):
        if (p < 0.0) or (p > 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p == 1.0):
            return float("inf")
        elif (p == 0.0):
            return float("-inf")
        else:
            low = -1.0
            high = 1.0
            while self.CDF(low) > p:
                low *= 2.0
            while self.CDF(high) < p:
                high *= 2.0
            diff = None
            while diff is None or abs(diff) > self.epsilon:
                mid = (low + high) / 2.0
                diff = self.CDF(mid) - p
                if diff > 0:
                    high = mid
                else:
                    low = mid
            return mid

            # # Using Newton-Raphson algorithm
            # if p <= .001:
            #     self.epsilon = 1e-5
            # p1 = p
            # if (p1 > 0.8) and (p1 < 0.9):
            #     p2 = .5
            # else:
            #     p2 = 0.85
            # counter = 0
            # while (abs(p1 - p2) > self.epsilon) or (counter < self.maxIter):
            #     q2 = (self.CDF(p2) - p)
            #     p1 = p2
            #     p2 = p1 - (q2/self.PDF(p1))
            #     counter += 1
            # return p2

################### Binomial
class BinomialDistribution(object):
    def __init__(self, size, p_success, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.prob = p_success
        self.n = float(size)
        if self.n < 0.0:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)
        elif (self.prob < 0.0) or (self.prob > 1.0):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        if x < 0.0:
            return 0.0
        else:
            if x == 0:
                nchoosek = 1.0
            elif (x < 0) or (x > self.n):
                nchoosek = 0.0
            else:
                nchoosek = nChooseK(self.n, x)
            return nchoosek * pow(self.prob, x) * pow(1.0 - self.prob, self.n - x)

    def CDF(self, x):
        if math.isnan(x):
            return float("nan")
        elif x < 0.0:
            return 0.0
        else:
            if (self.n - x <= 0.0) or (self.prob == 0.0):
                return 1.0
            else:
                x = math.floor(x)
                return incompleteBetaFunction(1.0 - self.prob, self.n - x, 1.0 + x)

    def QF(self, p):
        if (p < 0.0) or (p > 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif math.isnan(p):
            return float("nan")
        elif (p == 1.0):
            return self.n
        elif (p == 0.0):
            return 0.0
        elif (p > 0.0) and (p < 1.0):
            # step through CDFs until we find the right one
            x = 0
            p0 = 0.0
            while (p0 < p):
                p0 = p0 + self.PDF(x)
                x += 1
            return x - 1
        else:
            return 0.0

################### Uniform
class UniformDistribution(object):
    def __init__(self, minimum, maximum, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.mi = minimum
        self.ma = maximum
        if self.mi >= self.ma:
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        if (x < self.mi) or (x > self.ma):
            return 0.0
        elif (math.isnan(x)):
            return float("nan")
        else:
            return 1.0/(self.ma - self.mi)

    def CDF(self, x):
        if (x < self.mi):
            return 0.0
        elif (x > self.ma):
            return 1.0
        else:
            return (x - self.mi)/(self.ma - self.mi)

    def QF(self, p):
        if (p > 1.0) or (p < 0.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p > 0.0) or (p < 1.0):
            return self.mi + p*(self.ma - self.mi)
        elif (math.isnan(p)):
            return float("nan")
        else:
            return 0.0

################### Geometric
class GeometricDistribution(object):
    def __init__(self, p_success, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.prob = p_success
        if (self.prob < 0.0) or (self.prob > 1.0):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        if (x < 0.0):
            return 0.0
        else:
            return self.prob*pow(1.0 - self.prob, x)

    def CDF(self, x):
        if (x < 0.0):
            return 0.0
        else:
            x = math.floor(x)
            return 1.0 - pow(1.0 - self.prob, x + 1.0)

    def QF(self, p):
        if (p > 1.0) or (p < 0.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (math.isnan(p)):
            return float("nan")
        elif (p == 1.0):
            return float("inf")
        elif (p > 0.0) and (p < 1.0):
            if self.prob == 1.0:
                return 0.0
            else:
                return math.floor(math.log(1.0 - p)/math.log(1.0 - self.prob))
        else:
            return 0.0

################### Hypergeometric
class HypergeometricDistribution(object):
    def __init__(self, n_white, n_black, n_drawn, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.n_white = n_white
        self.n_black = n_black
        self.n_drawn = n_drawn
        if (n_white + n_black < n_drawn):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x): # x is number of white balls drawn
        # compute nchoosek(n_white,x)
        if x == 0:
            nchoosek1 = 1.0
        elif (x < 0) or (x >= self.n_white):
            nchoosek1 = 0.0
        else:
            nchoosek1 = nChooseK(self.n_white, x)
        # compute nchoosek(n_black, n_drawn - x)
        if (self.n_drawn - x == 0):
            nchoosek2 = 1.0
        elif (self.n_drawn - x < 0) or (self.n_drawn - x > self.n_black):
            nchoosek2 = 0.0
        else:
            nchoosek2 = nChooseK(self.n_black, self.n_drawn - x)
        # compute nchoosek(n_white + n_black, n_drawn)
        if self.n_drawn == 0:
            nchoosek3 = 1.0
        elif (self.n_drawn < 0) or (self.n_drawn > self.n_white + self.n_black):
            nchoosek3 = 0.0
        else:
            nchoosek3 = nChooseK(self.n_white + self.n_black, self.n_drawn)
        # compute
        return nchoosek1 * nchoosek2 / nchoosek3

    def CDF(self, x):
        if (x > self.n_white):
            return 0.0
        else:
            val = 0.0
            for i in range(0, int(math.floor(x + 1.0))):
                val = val + self.PDF(float(i))
            return val

    def QF(self, p):
        if (p > 1.0) or (p < 0.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif math.isnan(p):
            return float("nan")
        elif (p == 1):
            return self.n_drawn
        else:
            # step through CDFs until we find the right one
            x = 0
            p0 = 0.0
            while (p0 <= p):
                p0 = p0 + self.PDF(x)
                x += 1
            return x - 1

################### Weibull
class WeibullDistribution(object):
    def __init__(self, shape, scale, name, errcodeBase, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.a = float(shape)
        self.b = float(scale)
        if (self.a <= 0.0) or (self.b <= 0.0):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        if math.isnan(x):
            return float("nan")
        elif x == 0.0:
            if self.a < 1.0:
                return float("nan")
            elif self.a == 1.0:
                return 1.0/self.b
            else:
                return 0.0
        elif x >= 0.0:
            term1 = (self.a/self.b)
            term2 = pow(x/self.b, self.a - 1.0)
            term3 = math.exp(-pow(x/self.b, self.a))
            return term1 * term2 * term3

        else:
            return 0.0

    def CDF(self, x):
        if math.isnan(x):
            return float("nan")
        elif x >= 0.0:
            return 1.0 - math.exp(-pow(x/self.b, self.a))
        else:
            return 0.0

    def QF(self, p):
        if (p < 0.0) or (p > 1.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif (p == 1.0):
            return float("inf")
        else:
            try:
                return self.b * pow(-math.log(1.0 - p), 1.0/self.a)
            except OverflowError:
                return float("inf")

################### NegativeBinomial
class NegativeBinomialDistribution(object):
    def __init__(self, n, p, errcodeBase, name, pos):
        self.name = name
        self.errcodeBase = errcodeBase
        self.pos = pos
        self.n = n
        self.prob = p
        if (p > 1.0) or (p <= 0.0) or (self.n < 0.0):
            raise PFARuntimeException("invalid parameterization", self.errcodeBase + 0, self.name, self.pos)

    def PDF(self, x):
        if (math.isnan(x)):
            return float("nan")
        elif (self.n == 0.0) and (x == 0.0):
            return 1.0
        elif (self.n == 0.0) and (x != 0.0):
            return 0.0
        elif (self.prob == 1.0) and (x != 0.0):
            return 0.0
        elif (self.prob == 1.0) and (x == 0.0):
            return 1.0
        elif (x >= 0.0):
            if x == 0:
                val = 1.0
            elif x < 0:
                val = 0.0
            else:
                val = nChooseK(self.n + x - 1.0, x)
            return val * pow(1.0 - self.prob, x) * pow(self.prob, self.n)
        else:
            return 0.0

    def CDF(self, x):
        if math.isnan(x):
            return float("nan")
        elif (self.n == 0.0) and (x == 0.0):
            return 1.0
        val = 0.0
        for i in range(0, int(math.floor(x + 1.0))):
            val = val + self.PDF(float(i))
        return val

    def QF(self, p):
        # CDF SEEMS MORE ACCURATE NOW, REVISIT
        if math.isnan(p):
            return float("nan")
        elif (self.n == 0.0) and (x == 0.0):
            return 0.0
        elif (self.n == 0.0):
            return 0.0
        elif (p == 0.0):
            return 0.0
        elif (p == 1.0):
            return float("inf")
        elif (p > 1.0) or (p < 0.0):
            raise PFARuntimeException("invalid input", self.errcodeBase + 1, self.name, self.pos)
        elif self.prob == 1.0:
            return 0.0
        elif (p > 0.0) or (p < 1.0):
            # using Cornish-Fisher expansion
            Q = 1.0/self.prob
            P = (1.0 - self.prob)*Q
            mu = self.n * (1.0 - self.prob)/self.prob
            sigma = math.sqrt(self.n * P * Q)
            gamma = (Q + P)/sigma
            z = GaussianDistribution(0.0, 1.0, self.name, self.errcodeBase, self.pos).QF(p)

            # CF approximation gets us close
            w = math.ceil(mu + sigma * (z + gamma * (z*z - 1.0)/6.0))
            # use math.ceil (next term has a minus sign)

            # CDF seems mildly unstable for extreme values
            # cant use newton raphson. The way this computes CDF doesnt
            # seem to be accurate enough for these extreme cases

            # CDF
            # only use CD for very extreme values
            if w > 100000:
                return w
            else:  # do the step-through-CDF method
                x = 0.0
                s = 0.0
                while (s <= p):
                    s = s + self.PDF(x)
                    x += 1
                return x - 1



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from titus.pfaast import *
from titus.datatype import *
from titus.util import bytesToString

class TestAstToJson(unittest.TestCase):
    def testEngineConfig(self):
        self.assertEqual(
        json.loads(EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}).toJson(lineNumbers=False)),
        json.loads('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "action": [{"+": [2, 2]}]
}'''))

        self.assertEqual(
        json.loads(EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        {},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}).toJson(lineNumbers=False)),
        json.loads('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "begin": [{"+": [2, 2]}],
  "action": [{"+": [2, 2]}],
  "end": [{"+": [2, 2]}]
}'''))

        self.assertEqual(
        json.loads(EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        {"f": FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])},
        None,
        None,
        {},
        {},
        None,
        None,
        None,
        {},
        {}).toJson(lineNumbers=False)),
        json.loads('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "begin": [{"+": [2, 2]}],
  "action": [{"+": [2, 2]}],
  "end": [{"+": [2, 2]}],
  "fcns": {"f": {"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}}
}'''))

        self.assertEqual(
        json.loads(EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        {"f": FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])},
        None,
        None,
        {"private": Cell(AvroInt(), "0", False, False, CellPoolSource.EMBEDDED)},
        {},
        None,
        None,
        None,
        {},
        {}).toJson(lineNumbers=False)),
        json.loads('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "begin": [{"+": [2, 2]}],
  "action": [{"+": [2, 2]}],
  "end": [{"+": [2, 2]}],
  "fcns": {"f": {"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}},
  "cells":{"private":{"type":"int","init":0,"shared":false,"rollback":false}}
}'''))

        self.assertEqual(json.loads(EngineConfig(
        "test",
         Method.MAP,
         AvroInt(),
         AvroString(),
         [Call("+", [LiteralInt(2), LiteralInt(2)])],
         [Call("+", [LiteralInt(2), LiteralInt(2)])],
         [Call("+", [LiteralInt(2), LiteralInt(2)])],
         {"f": FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])},
         None,
         None,
         {"private": Cell(AvroInt(), "0", False, False, CellPoolSource.EMBEDDED)},
         {"private": Pool(AvroInt(), {}, False, False, CellPoolSource.EMBEDDED)},
         None,
         None,
         None,
         {},
         {}).toJson(lineNumbers=False)),
         json.loads('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "begin": [{"+": [2, 2]}],
  "action": [{"+": [2, 2]}],
  "end": [{"+": [2, 2]}],
  "fcns": {"f": {"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}},
  "cells":{"private":{"type":"int","init":0,"shared":false,"rollback":false}},
  "pools":{"private":{"type":"int","init":{},"shared":false,"rollback":false}}
}'''))

        self.assertEqual(json.loads(EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        {"f": FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])},
        None,
        None,
        {"private": Cell(AvroInt(), "0", False, False, CellPoolSource.EMBEDDED)},
        {"private": Pool(AvroInt(), {}, False, False, CellPoolSource.EMBEDDED)},
        12345,
        None,
        None,
        {},
        {}).toJson(lineNumbers=False)),
    json.loads('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "begin": [{"+": [2, 2]}],
  "action": [{"+": [2, 2]}],
  "end": [{"+": [2, 2]}],
  "fcns": {"f": {"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}},
  "cells":{"private":{"type":"int","init":0,"shared":false,"rollback":false}},
  "pools":{"private":{"type":"int","init":{},"shared":false,"rollback":false}},
  "randseed":12345
}'''))

        self.assertEqual(json.loads(EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        {"f": FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])},
        None,
        None,
        {"private": Cell(AvroInt(), "0", False, False, CellPoolSource.EMBEDDED)},
        {"private": Pool(AvroInt(), {}, False, False, CellPoolSource.EMBEDDED)},
        12345,
        "hello",
        None,
        {},
        {}).toJson(lineNumbers=False)),
        json.loads('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "begin": [{"+": [2, 2]}],
  "action": [{"+": [2, 2]}],
  "end": [{"+": [2, 2]}],
  "fcns": {"f": {"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}},
  "cells":{"private":{"type":"int","init":0,"shared":false,"rollback":false}},
  "pools":{"private":{"type":"int","init":{},"shared":false,"rollback":false}},
  "randseed":12345,
  "doc":"hello"
}'''))

        self.assertEqual(json.loads(EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        {"f": FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])},
        None,
        None,
        {"private": Cell(AvroInt(), "0", False, False, CellPoolSource.EMBEDDED)},
        {"private": Pool(AvroInt(), {}, False, False, CellPoolSource.EMBEDDED)},
        12345,
        "hello",
        None,
        {"internal": "data"},
        {}).toJson(lineNumbers=False)),
        json.loads('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "begin": [{"+": [2, 2]}],
  "action": [{"+": [2, 2]}],
  "end": [{"+": [2, 2]}],
  "fcns": {"f": {"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}},
  "cells":{"private":{"type":"int","init":0,"shared":false,"rollback":false}},
  "pools":{"private":{"type":"int","init":{},"shared":false,"rollback":false}},
  "randseed":12345,
  "doc":"hello",
  "metadata":{"internal":"data"}
}'''))

        self.assertEqual(json.loads(EngineConfig(
        "test",
        Method.MAP,
        AvroInt(),
        AvroString(),
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        [Call("+", [LiteralInt(2), LiteralInt(2)])],
        {"f": FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])},
        None,
        None,
        {"private": Cell(AvroInt(), "0", False, False, CellPoolSource.EMBEDDED)},
        {"private": Pool(AvroInt(), {}, False, False, CellPoolSource.EMBEDDED)},
        12345,
        "hello",
        None,
        {"internal": "data"},
        {"param": json.loads("3")}).toJson(lineNumbers=False)),
        json.loads('''{
  "name": "test",
  "method": "map",
  "input": "int",
  "output": "string",
  "begin": [{"+": [2, 2]}],
  "action": [{"+": [2, 2]}],
  "end": [{"+": [2, 2]}],
  "fcns": {"f": {"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}},
  "cells":{"private":{"type":"int","init":0,"shared":false,"rollback":false}},
  "pools":{"private":{"type":"int","init":{},"shared":false,"rollback":false}},
  "randseed":12345,
  "doc":"hello",
  "metadata":{"internal":"data"},
  "options":{"param":3}
}'''))

    def testCell(self):
        self.assertEqual(json.loads(Cell(AvroInt(), "0", False, False, CellPoolSource.EMBEDDED).toJson(lineNumbers=False)), json.loads('''{"type":"int","init":0,"shared":false,"rollback":false}'''))

    def testPool(self):
        self.assertEqual(json.loads(Pool(AvroInt(), {}, False, False, CellPoolSource.EMBEDDED).toJson(lineNumbers=False)), json.loads('''{"type":"int","init":{},"shared":false,"rollback":false}'''))
        self.assertEqual(json.loads(Pool(AvroInt(), {"one": "1"}, False, False, CellPoolSource.EMBEDDED).toJson(lineNumbers=False)), json.loads('''{"type":"int","init":{"one":1},"shared":false,"rollback":false}'''))

    def testDefineFunction(self):
        self.assertEqual(json.loads(FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()]).toJson(lineNumbers=False)), json.loads('''{"params":[{"x":"int"},{"y":"string"}],"ret":"null","do":[null]}'''))

    def testCallFunction(self):
        self.assertEqual(json.loads(Call("+", [LiteralInt(2), LiteralInt(2)]).toJson(lineNumbers=False)), json.loads('''{"+":[2,2]}'''))

    def testCallWithFcn(self):
        self.assertEqual(json.loads(Call("sort", [Ref("array"), FcnRef("byname")]).toJson(lineNumbers=False)), json.loads('''{"sort":["array",{"fcn": "byname"}]}'''))

    def testCallWithFcnDef(self):
        self.assertEqual(json.loads(Call("sort", [Ref("array"), FcnDef([{"x": AvroInt()}, {"y": AvroString()}], AvroNull(), [LiteralNull()])]).toJson(lineNumbers=False)),
                         json.loads('''{"sort":["array",{"params": [{"x": "int"}, {"y": "string"}], "ret": "null", "do": [null]}]}'''))

    def testRef(self):
        self.assertEqual(json.loads(Ref("x").toJson(lineNumbers=False)), "x")

    def testNull(self):
        self.assertEqual(json.loads(LiteralNull().toJson(lineNumbers=False)), json.loads("null"))

    def testBoolean(self):
        self.assertEqual(json.loads(LiteralBoolean(True).toJson(lineNumbers=False)), json.loads("true"))
        self.assertEqual(json.loads(LiteralBoolean(False).toJson(lineNumbers=False)), json.loads("false"))

    def testInt(self):
        self.assertEqual(json.loads(LiteralInt(2).toJson(lineNumbers=False)), json.loads('''2'''))

    def testLong(self):
        self.assertEqual(json.loads(LiteralLong(2).toJson(lineNumbers=False)), json.loads('''{"long":2}'''))

    def testFloat(self):
        self.assertEqual(json.loads(LiteralFloat(2.5).toJson(lineNumbers=False)), json.loads('''{"float":2.5}'''))

    def testDouble(self):
        self.assertEqual(json.loads(LiteralDouble(2.2).toJson(lineNumbers=False)), json.loads("2.2"))

    def testString(self):
        self.assertEqual(json.loads(LiteralString("hello").toJson(lineNumbers=False)), json.loads('''{"string":"hello"}'''))

    def testBase64(self):
        self.assertEqual(json.loads(LiteralBase64(bytesToString("hello".encode('utf-8'))).toJson(lineNumbers=False)), json.loads('''{"base64":"aGVsbG8="}'''))
        self.assertEqual(json.loads(LiteralBase64(bytesToString('Δ'.encode('utf-8'))).toJson(lineNumbers=False)), json.loads('''{"base64":"zpQ="}'''))

    def testLiteral(self):
        self.assertEqual(json.loads(Literal(AvroRecord([AvroField("one", AvroInt()), AvroField("two", AvroDouble()), AvroField("three", AvroString())], "SimpleRecord"), '''{"one": 1, "two": 2.2, "three": "THREE"}''').toJson(lineNumbers=False)),
        json.loads('''{"type":{"type":"record","name":"SimpleRecord","fields":[{"name":"one","type":"int"},{"name":"two","type":"double"},{"name":"three","type":"string"}]},"value":{"one":1,"two":2.2,"three":"THREE"}}'''))

    def testNewRecord(self):
        self.assertEqual(json.loads(NewObject({"one": LiteralInt(1), "two": LiteralDouble(2.2), "three": LiteralString("THREE")},
                                   AvroRecord([AvroField("one", AvroInt()), AvroField("two", AvroDouble()), AvroField("three", AvroString())], "SimpleRecord"), AvroTypeBuilder()).toJson(lineNumbers=False)),
        json.loads('''{"new":{"one":1,"two":2.2,"three":{"string":"THREE"}},"type":{"type":"record","name":"SimpleRecord","fields":[{"name":"one","type":"int"},{"name":"two","type":"double"},{"name":"three","type":"string"}]}}'''))

    def testNewArray(self):
        self.assertEqual(json.loads(NewArray([LiteralInt(1), LiteralInt(2), LiteralInt(3)], AvroArray(AvroInt()), AvroTypeBuilder()).toJson(lineNumbers=False)), json.loads('''{"new":[1,2,3],"type":{"type":"array","items":"int"}}'''))

    def testDo(self):
        self.assertEqual(json.loads(Do([Ref("x"), Ref("y"), Ref("z")]).toJson(lineNumbers=False)), json.loads('''{"do":["x","y","z"]}'''))

    def testLet(self):
        self.assertEqual(json.loads(Let({"x": LiteralInt(3), "y": LiteralInt(4)}).toJson(lineNumbers=False)), json.loads('''{"let":{"x":3,"y":4}}'''))

    def testSet(self):
        self.assertEqual(json.loads(SetVar({"x": LiteralInt(3), "y": LiteralInt(4)}).toJson(lineNumbers=False)), json.loads('''{"set":{"x":3,"y":4}}'''))

    def testAttrGet(self):
        self.assertEqual(json.loads(AttrGet(Ref("a"), [Ref("a"), LiteralInt(1), LiteralString("b")]).toJson(lineNumbers=False)), json.loads('''{"attr":"a","path":["a",1,{"string":"b"}]}'''))

    def testAttrSet(self):
        self.assertEqual(json.loads(AttrTo(Ref("a"), [Ref("a"), LiteralInt(1), LiteralString("b")], LiteralDouble(2.2)).toJson(lineNumbers=False)), json.loads('''{"attr":"a","path":["a",1,{"string":"b"}],"to":2.2}'''))

    def testCellGet(self):
        self.assertEqual(json.loads(CellGet("c", []).toJson(lineNumbers=False)), json.loads('''{"cell":"c"}'''))
        self.assertEqual(json.loads(CellGet("c", [Ref("a"), LiteralInt(1), LiteralString("b")]).toJson(lineNumbers=False)), json.loads('''{"cell":"c","path":["a",1,{"string":"b"}]}'''))

    def testCellSet(self):
        self.assertEqual(json.loads(CellTo("c", [], LiteralDouble(2.2)).toJson(lineNumbers=False)), json.loads('''{"cell":"c","to":2.2}'''))
        self.assertEqual(json.loads(CellTo("c", [Ref("a"), LiteralInt(1), LiteralString("b")], LiteralDouble(2.2)).toJson(lineNumbers=False)), json.loads('''{"cell":"c","path":["a",1,{"string":"b"}],"to":2.2}'''))

    def testPoolGet(self):
        self.assertEqual(json.loads(PoolGet("p", [Ref("a"), LiteralInt(1), LiteralString("b")]).toJson(lineNumbers=False)), json.loads('''{"pool":"p","path":["a",1,{"string":"b"}]}'''))

    def testPoolSet(self):
        self.assertEqual(json.loads(PoolTo("p", [Ref("a"), LiteralInt(1), LiteralString("b")], LiteralDouble(2.2), LiteralDouble(2.2)).toJson(lineNumbers=False)), json.loads('''{"pool":"p","path":["a",1,{"string":"b"}],"to":2.2,"init":2.2}'''))

    def testIf(self):
        self.assertEqual(json.loads(If(LiteralBoolean(True), [Ref("x")], None).toJson(lineNumbers=False)), json.loads('''{"if":true,"then":["x"]}'''))
        self.assertEqual(json.loads(If(LiteralBoolean(True), [Ref("x")], [Ref("y")]).toJson(lineNumbers=False)), json.loads('''{"if":true,"then":["x"],"else":["y"]}'''))

    def testCond(self):
        self.assertEqual(json.loads(Cond([If(LiteralBoolean(False), [Ref("x")], None), If(LiteralBoolean(True), [Ref("y")], None)], None).toJson(lineNumbers=False)),
                         json.loads('''{"cond":[{"if":false,"then":["x"]},{"if":true,"then":["y"]}]}'''))
        self.assertEqual(json.loads(Cond([If(LiteralBoolean(False), [Ref("x")], None), If(LiteralBoolean(True), [Ref("y")], None)], [Ref("z")]).toJson(lineNumbers=False)),
                         json.loads('''{"cond":[{"if":false,"then":["x"]},{"if":true,"then":["y"]}],"else":["z"]}'''))

    def testWhile(self):
        self.assertEqual(json.loads(While(LiteralBoolean(True), [Call("+", [LiteralInt(2), LiteralInt(2)])]).toJson(lineNumbers=False)),
                         json.loads('''{"while":true,"do":[{"+":[2,2]}]}'''))

    def testDoUntil(self):
        self.assertEqual(json.loads(DoUntil([Call("+", [LiteralInt(2), LiteralInt(2)])], LiteralBoolean(True)).toJson(lineNumbers=False)),
                         json.loads('''{"do":[{"+":[2,2]}],"until":true}'''))

    def testFor(self):
        self.assertEqual(json.loads(For({"i": LiteralInt(0)}, Call("<", [Ref("i"), LiteralInt(10)]), {"i": Call("+", [Ref("i"), LiteralInt(1)])}, [Ref("i")], False).toJson(lineNumbers=False)),
                         json.loads('''{"for":{"i":0},"while":{"<":["i",10]},"step":{"i":{"+":["i",1]}},"do":["i"]}'''))

    def testForeach(self):
        self.assertEqual(json.loads(Foreach("x", Literal(AvroArray(AvroInt()), '''[1, 2, 3]'''), [Ref("x")], False).toJson(lineNumbers=False)),
                         json.loads('''{"foreach":"x","in":{"type":{"type":"array","items":"int"},"value":[1,2,3]},"do":["x"],"seq":false}'''))

    def testForkeyval(self):
        self.assertEqual(json.loads(Forkeyval("k", "v", Literal(AvroMap(AvroInt()), '''{"one": 1, "two": 2, "three": 3}'''), [Ref("k")]).toJson(lineNumbers=False)),
                         json.loads('''{"forkey":"k","forval":"v","in":{"type":{"type":"map","values":"int"},"value":{"one":1,"two":2,"three":3}},"do":["k"]}'''))

    def testCast(self):
        self.assertEqual(json.loads(CastBlock(LiteralInt(3), [CastCase(AvroString(), "x", [Ref("x")]), CastCase(AvroInt(), "x", [Ref("x")])], False).toJson(lineNumbers=False)),
                         json.loads('''{"cast":3,"cases":[{"as":"string","named":"x","do":["x"]},{"as":"int","named":"x","do":["x"]}],"partial":false}'''))

    def testDoc(self):
        self.assertEqual(json.loads(Doc("hello").toJson(lineNumbers=False)), json.loads('''{"doc":"hello"}'''))

    def testError(self):
        self.assertEqual(json.loads(Error("hello", None).toJson(lineNumbers=False)), json.loads('''{"error":"hello"}'''))
        self.assertEqual(json.loads(Error("hello", -3).toJson(lineNumbers=False)), json.loads('''{"error":"hello","code":-3}'''))

    def testLog(self):
        self.assertEqual(json.loads(Log([LiteralString("hello")], None).toJson(lineNumbers=False)), json.loads('''{"log":[{"string":"hello"}]}'''))
        self.assertEqual(json.loads(Log([LiteralString("hello")], "DEBUG").toJson(lineNumbers=False)), json.loads('''{"log":[{"string":"hello"}],"namespace":"DEBUG"}'''))
        self.assertEqual(json.loads(Log([Call("+", [LiteralInt(2), LiteralInt(2)])], None).toJson(lineNumbers=False)), json.loads('''{"log":[{"+":[2,2]}]}'''))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from titus.datatype import Type
from titus.datatype import FcnType
from titus.datatype import ExceptionType
from titus.datatype import AvroType
from titus.datatype import AvroCompiled
from titus.datatype import AvroNull
from titus.datatype import AvroBoolean
from titus.datatype import AvroInt
from titus.datatype import AvroLong
from titus.datatype import AvroFloat
from titus.datatype import AvroDouble
from titus.datatype import AvroBytes
from titus.datatype import AvroFixed
from titus.datatype import AvroString
from titus.datatype import AvroEnum
from titus.datatype import AvroArray
from titus.datatype import AvroMap
from titus.datatype import AvroRecord
from titus.datatype import AvroField
from titus.datatype import AvroUnion

class Pattern(object):
    """Trait for a type pattern.

    A type pattern is something that a titus.datatype.AvroType is matched against when determining if a PFA function signature can be applied to a given set of arguments.

    It could be as simple as the type itself (e.g. P.Int() matches AvroInt()) or it could be a complex wildcard.
    """
    pass

class Null(Pattern):
    """Matches titus.datatype.AvroNull."""
    def __repr__(self):
        return "P.Null()"

class Boolean(Pattern):
    """Matches titus.datatype.AvroBoolean."""
    def __repr__(self):
        return "P.Boolean()"

class Int(Pattern):
    """Matches titus.datatype.AvroInt."""
    def __repr__(self):
        return "P.Int()"

class Long(Pattern):
    """Matches titus.datatype.AvroLong."""
    def __repr__(self):
        return "P.Long()"

class Float(Pattern):
    """Matches titus.datatype.AvroFloat."""
    def __repr__(self):
        return "P.Float()"

class Double(Pattern):
    """Matches titus.datatype.AvroDouble."""
    def __repr__(self):
        return "P.Double()"

class Bytes(Pattern):
    """Matches titus.datatype.AvroBytes."""
    def __repr__(self):
        return "P.Bytes()"

class String(Pattern):
    """Matches titus.datatype.AvroString."""
    def __repr__(self):
        return "P.String()"

class Array(Pattern):
    """Matches titus.datatype.AvroArray with a given items pattern."""
    def __init__(self, items):
        """:type items: titus.P
        :param items: pattern for the array items
        """
        self._items = items
    @property
    def items(self):
        return self._items
    def __repr__(self):
        return "P.Array(" + repr(self.items) + ")"

class Map(Pattern):
    """Matches titus.datatype.AvroMap with a given values pattern."""
    def __init__(self, values):
        """:type values: titus.P
        :param values: pattern for the map values
        """
        self._values = values
    @property
    def values(self):
        return self._values
    def __repr__(self):
        return "P.Map(" + repr(self.values) + ")"

class Union(Pattern):
    """Matches titus.datatype.AvroUnion with a given set of sub-patterns."""
    def __init__(self, types):
        """:type types: list of titus.P
        :param types: patterns for the subtypes
        """
        self._types = types
    @property
    def types(self):
        return self._types
    def __repr__(self):
        return "P.Union(" + repr(self.types) + ")"

class Fixed(Pattern):
    """Matches titus.datatype.AvroFixed with a given size and an optional name.
    
    Note: not used for any PFA patterns, and probably shouldn't be, even in the future. titus.P.WildFixed is used instead.
    """
    def __init__(self, size, fullName=None):
        """:type size: positive integer
        :param size: width of the fixed-width byte array
        :type fullName: string or ``None``
        :param fullName: optional name of the fixed pattern (if not provided, fixed types of any name would match)
        """
        self._size = size
        self._fullName = fullName
    @property
    def size(self):
        return self._size
    @property
    def fullName(self):
        return self._fullName
    def __repr__(self):
        return "P.Fixed(" + repr(self.size) + ", " + repr(self.fullName) + ")"

class Enum(Pattern):
    """Matches titus.datatype.AvroEnum with a given symbols and an optional name.
    
    Note: not used for any PFA patterns, and probably shouldn't be, even in the future. titus.P.WildEnum is used instead.
    """
    def __init__(self, symbols, fullName=None):
        """:type symbols: list of strings
        :param symbols: names of the enum symbols
        :type fullName: string or ``None``
        :param fullName: optional name of the enum pattern (if not provided, enum types of any name would match)
        """
        self._symbols = tuple(symbols)
        self._fullName = fullName
    @property
    def symbols(self):
        return self._symbols
    @property
    def fullName(self):
        return self._fullName
    def __repr__(self):
        return "P.Enum(" + repr(self.symbols) + ", " + repr(self.fullName) + ")"

class Record(Pattern):
    """Matches titus.datatype.AvroRecord with given fields and an optional name.
    
    Note: not used for any PFA patterns, and probably shouldn't be, even in the future. titus.P.WildRecord is used instead.
    """
    def __init__(self, fields, fullName=None):
        """:type fields: dict from field names to titus.P
        :param fields: patterns for the record fields
        :type fullName: string or ``None``
        :param fullName: optional name of the record pattern (if not provided, record types of any name would match)
        """
        self._fields = fields
        self._fullName = fullName
    @property
    def fields(self):
        return self._fields
    @property
    def fullName(self):
        return self._fullName
    def __repr__(self):
        return "P.Record(" + repr(self.fields) + ", " + repr(self.fullName) + ")"

class Fcn(Pattern):
    """Matches titus.datatype.FcnType with a given sequence of parameter patterns and return pattern."""
    def __init__(self, params, ret):
        """:type params: list of titus.P
        :param params: patterns for each of the slots in the function-argument's signature (must have only one signature with no wildcards)
        :type ret: titus.P
        :param ret: pattern for the return type of the function-argument
        """
        self._params = params
        self._ret = ret
    @property
    def params(self):
        return self._params
    @property
    def ret(self):
        return self._ret
    def __repr__(self):
        return "P.Fcn(" + repr(self.params) + ", " + repr(self.ret) + ")"

class Wildcard(Pattern):
    """Matches any titus.datatype.AvroType or one of a restricted set.

    Label letters are shared across a signature (e.g. if two wildcards are both labeled "A", then they both have to resolve to the same type).
    """
    def __init__(self, label, oneOf=None):
        """:type label: string
        :param label: label letter (usually one character long, but in principle an arbitrary string)
        :type oneOf: list of titus.datatype.AvroType or ``None``
        :param oneOf: allowed types or ``None`` for unrestricted
        """
        self._label = label
        self._oneOf = oneOf
    @property
    def label(self):
        return self._label
    @property
    def oneOf(self):
        return self._oneOf
    def __repr__(self):
        return "P.Wildcard(" + repr(self.label) + ", " + repr(self.oneOf) + ")"

class WildRecord(Pattern):
    """Matches a titus.datatype.AvroRecord with *at least* the requested set of fields.

    Label letters are shared across a signature (e.g. if two wildcards are both labeled "A", then they both have to resolve to the same type).
    """
    def __init__(self, label, minimalFields=None):
        """:type label: string
        :param label: label letter (usually one character long, but in principle an arbitrary string)
        :type minimalFields: dict from field name to titus.P
        :param minimalFields: fields that a matching record must have and patterns for their respective types
        """
        self._label = label
        self._minimalFields = minimalFields
    @property
    def label(self):
        return self._label
    @property
    def minimalFields(self):
        return self._minimalFields
    def __repr__(self):
        return "P.WildRecord(" + repr(self.label) + ", " + repr(self.minimalFields) + ")"

class WildEnum(Pattern):
    """Matches a titus.datatype.AvroEnum without any constraint on the symbol names.

    Label letters are shared across a signature (e.g. if two wildcards are both labeled "A", then they both have to resolve to the same type).
    """
    def __init__(self, label):
        """:type label: string
        :param label: label letter (usually one character long, but in principle an arbitrary string)
        """
        self._label = label
    @property
    def label(self):
        return self._label
    def __repr__(self):
        return "P.WildEnum(" + repr(self.label) + ")"

class WildFixed(Pattern):
    """Matches a titus.datatype.AvroFixed without any constraint on the size.

    Label letters are shared across a signature (e.g. if two wildcards are both labeled "A", then they both have to resolve to the same type).
    """
    def __init__(self, label):
        """:type label: string
        :param label: label letter (usually one character long, but in principle an arbitrary string)
        """
        self._label = label
    @property
    def label(self):
        return self._label
    def __repr__(self):
        return "P.WildFixed(" + repr(self.label) + ")"

class EnumFields(Pattern):
    """Matches a titus.datatype.AvroEnum whose symbols match the fields of a given record

    Label letters are shared across a signature (e.g. if two wildcards are both labeled "A", then they both have to resolve to the same type).
    """
    def __init__(self, label, wildRecord):
        """:type label: string
        :param label: label letter (usually one character long, but in principle an arbitrary string)
        :type wildRecord: string
        :param wildRecord: label letter of the record (also a wildcard)
        """
        self._label = label
        self._wildRecord = wildRecord
    @property
    def label(self):
        return self._label
    @property
    def wildRecord(self):
        return self._wildRecord
    def __repr__(self):
        return "P.EnumFields(" + repr(self.label) + ", " + repr(self.wildRecord) + ")"

def toType(pat):
    """Convert a pattern to a type, if possible (wildcards can't be converted to types).

    :type pat: titus.P
    :param pat: pattern to convert
    :rtype: titus.datatype.Type
    :return: corresponding type (titus.datatype.Type rather than titus.datatype.AvroType to allow for titus.datatype.FcnType)
    """

    if isinstance(pat, Null): return AvroNull()
    elif isinstance(pat, Boolean): return AvroBoolean()
    elif isinstance(pat, Int): return AvroInt()
    elif isinstance(pat, Long): return AvroLong()
    elif isinstance(pat, Float): return AvroFloat()
    elif isinstance(pat, Double): return AvroDouble()
    elif isinstance(pat, Bytes): return AvroBytes()
    elif isinstance(pat, String): return AvroString()

    elif isinstance(pat, Array): return AvroArray(toType(pat.items))
    elif isinstance(pat, Map): return AvroMap(toType(pat.values))
    elif isinstance(pat, Union): return AvroUnion([toType(x) for x in pat.types])

    elif isinstance(pat, Fixed) and pat.fullName is not None:
        namebits = pat.fullName.split(".")
        if len(namebits) == 1:
            return AvroFixed(pat.size, namebits[-1], None)
        else:
            return AvroFixed(pat.size, namebits[-1], ".".join(namebits[:-1]))
    elif isinstance(pat, Fixed):
        return AvroFixed(pat.size)

    elif isinstance(pat, Enum) and pat.fullName is not None:
        namebits = pat.fullName.split(".")
        if len(namebits) == 1:
            return AvroEnum(pat.symbols, namebits[-1], None)
        else:
            return AvroEnum(pat.symbols, namebits[-1], ".".join(namebits[:-1]))
    elif isinstance(pat, Enum):
        return AvroEnum(pat.symbols)

    elif isinstance(pat, Record) and pat.fullName is not None:
        namebits = pat.fullName.split(".")
        if len(namebits) == 1:
            return AvroRecord([AvroField(k, toType(v)) for k, v in list(pat.fields.items())], namebits[-1], None)
        else:
            return AvroRecord([AvroField(k, toType(v)) for k, v in list(pat.fields.items())], namebits[-1], ".".join(namebits[:-1]))
    elif isinstance(pat, Record):
        return AvroRecord([AvroField(k, toType(v)) for k, v in list(pat.fields.items())])

    elif isinstance(pat, Fcn): return FcnType([toType(x) for x in pat.params()], toType(pat.ret()))

    else: raise Exception

def fromType(t, memo=None):
    """Convert a type to a pattern.

    :type t: titus.datatype.AvroType
    :param t: type to convert
    :rtype: titus.P
    :return: corresponding pattern
    """

    if memo is None:
        memo = set()

    if isinstance(t, AvroNull): return Null()
    elif isinstance(t, AvroBoolean): return Boolean()
    elif isinstance(t, AvroInt): return Int()
    elif isinstance(t, AvroLong): return Long()
    elif isinstance(t, AvroFloat): return Float()
    elif isinstance(t, AvroDouble): return Double()
    elif isinstance(t, AvroBytes): return Bytes()
    elif isinstance(t, AvroString): return String()

    elif isinstance(t, AvroArray): return Array(fromType(t.items, memo))
    elif isinstance(t, AvroMap): return Map(fromType(t.values, memo))
    elif isinstance(t, AvroUnion): return Union([fromType(x, memo) for x in t.types])

    elif isinstance(t, AvroFixed) and t.namespace is not None and t.namespace != "": return Fixed(t.size, t.namespace + "." + t.name)
    elif isinstance(t, AvroFixed): return Fixed(t.size, t.name)
    elif isinstance(t, AvroEnum) and t.namespace is not None and t.namespace != "": return Enum(t.symbols, t.namespace + "." + t.name)
    elif isinstance(t, AvroEnum): return Enum(t.symbols, t.name)
    elif isinstance(t, AvroRecord) and t.namespace is not None and t.namespace != "":
        name = t.namespace + "." + t.name
        if name in memo:
            return Record({}, name)
        else:
            return Record(dict((f.name, fromType(f.avroType, memo.union(set([name])))) for f in t.fields), name)
    elif isinstance(t, AvroRecord):
        if t.name in memo:
            return Record({}, t.name)
        else:
            return Record(dict((f.name, fromType(f.avroType, memo.union(set([t.name])))) for f in t.fields), t.name)
    elif isinstance(t, FcnType): return Fcn([fromType(x, memo) for x in t.params()], fromType(t.ret(), memo))
    elif isinstance(t, ExceptionType): raise IncompatibleTypes("exception type cannot be used in argument patterns")

def mustBeAvro(t):
    """Raise a ``TypeError`` if a given titus.datatype.Type is not a titus.datatype.AvroType; otherwise, pass through.

    :type t: titus.datatype.Type
    :param t: type to check
    :rtype: titus.datatype.AvroType
    :return: the input ``t`` or raise an exception
    """
    if not isinstance(t, AvroType):
        raise TypeError(repr(t) + " is not an Avro type")
    else:
        return t



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn, div
import titus.P as P
from titus.lib.array import argLowestN
from titus.lib.prob.dist import Chi2Distribution

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "m.link."

#################################################################### 

class SoftMax(LibFcn):
    name = prefix + "softmax"
    sig  = Sigs([Sig([{"x": P.Array(P.Double())}], P.Array(P.Double())),
                 Sig([{"x": P.Map(P.Double())}], P.Map(P.Double()))])
    errcodeBase = 25000
    def __call__(self, state, scope, pos, paramTypes, x):
        if len(x) == 0:
            raise PFARuntimeException("empty input", self.errcodeBase + 0, self.name, pos)
        if paramTypes[0]["type"] == "map":
            xx = x.copy()
            tmp = list(map(abs, list(xx.values())))
            if list(xx.values())[tmp.index(max(tmp))] >= 0:
                m = max(xx.values())
            else:
                m = min(xx.values())
            denom = sum([math.exp(v - m) for v in list(x.values())])
            for key in list(x.keys()):
                xx[key] = float(math.exp(xx[key] - m)/denom)
            return xx
        else:
            tmp = list(map(abs, x))
            if x[tmp.index(max(tmp))] >= 0:
                m = max(x)
            else:
                m = min(x)
            denom = sum([math.exp(v - m) for v in x])
            return [float(math.exp(val - m)/denom) for val in x]
provide(SoftMax())

def unwrapForNorm(x, func):
    if isinstance(x, dict):
        xx = x.copy()
        for key, val in zip(list(x.keys()), list(x.values())):
            xx[key] = float(func(val))
        return xx
    elif isinstance(x, (tuple, list)):
        xx = x[:]
        for i, val in enumerate(x):
            xx[i] = float(func(val))
        return xx
    else:
        return float(func(x))

class Logit(LibFcn):
    name = prefix + "logit"
    sig  = Sigs([Sig([{"x": P.Array(P.Double())}], P.Array(P.Double())),
                 Sig([{"x": P.Map(P.Double())}], P.Map(P.Double())),
                 Sig([{"x": P.Double()}], P.Double())])
    errcodeBase = 25010
    def __call__(self, state, scope, pos, paramTypes, x):
        return unwrapForNorm(x, lambda y: 1./(1. + math.exp(-y)))
provide(Logit())

class Probit(LibFcn):
    name = prefix + "probit"
    sig  = Sigs([Sig([{"x": P.Array(P.Double())}], P.Array(P.Double())),
                 Sig([{"x": P.Map(P.Double())}], P.Map(P.Double())),
                 Sig([{"x": P.Double()}], P.Double())])
    errcodeBase = 25020
    def __call__(self, state, scope, pos, paramTypes, x):
        return unwrapForNorm(x, lambda y: (math.erf(y/math.sqrt(2.)) + 1.)/2.)
provide(Probit())

class CLoglog(LibFcn):
    name = prefix + "cloglog"
    sig  = Sigs([Sig([{"x": P.Array(P.Double())}], P.Array(P.Double())),
                 Sig([{"x": P.Map(P.Double())}], P.Map(P.Double())),
                 Sig([{"x": P.Double()}], P.Double())])
    errcodeBase = 25030
    def __call__(self, state, scope, pos, paramTypes, x):
        return unwrapForNorm(x, lambda y: 1. - math.exp(-math.exp(y)))
provide(CLoglog())

class LogLog(LibFcn):
    name = prefix + "loglog"
    sig  = Sigs([Sig([{"x": P.Array(P.Double())}], P.Array(P.Double())),
                 Sig([{"x": P.Map(P.Double())}], P.Map(P.Double())),
                 Sig([{"x": P.Double()}], P.Double())])
    errcodeBase = 25040
    def __call__(self, state, scope, pos, paramTypes, x):
        return unwrapForNorm(x, lambda y: math.exp(-math.exp(y)))
provide(LogLog())

class Cauchit(LibFcn):
    name = prefix + "cauchit"
    sig  = Sigs([Sig([{"x": P.Array(P.Double())}], P.Array(P.Double())),
                 Sig([{"x": P.Map(P.Double())}], P.Map(P.Double())),
                 Sig([{"x": P.Double()}], P.Double())])
    errcodeBase = 25050
    def __call__(self, state, scope, pos, paramTypes, x):
        return unwrapForNorm(x, lambda y: 0.5 + (1./math.pi)*math.atan(y))
provide(Cauchit())

class Softplus(LibFcn):
    name = prefix + "softplus"
    sig  = Sigs([Sig([{"x": P.Array(P.Double())}], P.Array(P.Double())),
                 Sig([{"x": P.Map(P.Double())}], P.Map(P.Double())),
                 Sig([{"x": P.Double()}], P.Double())])
    errcodeBase = 25060
    def __call__(self, state, scope, pos, paramTypes, x):
        return unwrapForNorm(x, lambda y: math.log(1.0 + math.exp(y)))
provide(Softplus())

class ReLu(LibFcn):
    name = prefix + "relu"
    sig  = Sigs([Sig([{"x": P.Array(P.Double())}], P.Array(P.Double())),
                 Sig([{"x": P.Map(P.Double())}], P.Map(P.Double())),
                 Sig([{"x": P.Double()}], P.Double())])
    errcodeBase = 25070
    def __call__(self, state, scope, pos, paramTypes, x):
        return unwrapForNorm(x, lambda y: float("nan") if math.isnan(y) else max(0, y))
provide(ReLu())

class Tanh(LibFcn):
    name = prefix + "tanh"
    sig  = Sigs([Sig([{"x": P.Array(P.Double())}], P.Array(P.Double())),
                 Sig([{"x": P.Map(P.Double())}], P.Map(P.Double())),
                 Sig([{"x": P.Double()}], P.Double())])
    errcodeBase = 25080
    def __call__(self, state, scope, pos, paramTypes, x):
        return unwrapForNorm(x, lambda y: math.tanh(y))
provide(Tanh())








